unit EditBitmapUtils;

{$mode objfpc}{$H+}

interface

uses
  Classes, Graphics, Windows, LazUTF8Classes;

type
  { TBitmap32 }
  TBitmap32 = class
  private
    FHandle: HBITMAP;
    FCanvas: TCanvas;
    FBits: PRGBQUAD;
    FSize: TSize;
    function GetBits: PRGBQUAD; inline;
    function GetHeight: Integer; inline;
    function GetRect: Classes.TRect; inline;
    function GetScanLine(Row: Integer): Windows.PRGBQUAD; inline;
    function GetWidth: Integer; inline;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure CreateNew(AWidth, AHeight: Integer; AColor: TColor = clBlack);
    function Duplicate: TBitmap32;
    procedure PaintTo(DstCanvas: TCanvas; DstRect: Classes.TRect);
    procedure PaintFrom(SrcCanvas: TCanvas; SrcRect: Classes.TRect);
    procedure SaveAsJPGStream(Stream: TStream; CompressionQuality: TJPEGQualityRange);
    procedure SaveAsJPGFile(FileName: string; CompressionQuality: TJPEGQualityRange);
    procedure LoadFromStream(Stream: TStream);
    procedure LoadFromFile(FileName: string);
    function GetPixel(X, Y: Integer): PRGBQUAD; inline;
    procedure SetPixel(X, Y: Integer; R, G, B: BYTE); inline;
    function NotEmpty: Boolean; inline;
  public
    property Handle: HBITMAP read FHandle;
    property Size: TSize read FSize;
    property Rect: Classes.TRect read GetRect;
    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
    property Canvas: TCanvas read FCanvas;
    property Bits: PRGBQUAD read GetBits;
    property ScanLine[Row: Integer]: PRGBQUAD read GetScanLine;
    property Pixel[X, Y: Integer]: PRGBQUAD read GetPixel;
  end;

implementation

{ TBitmap32 }

function TBitmap32.GetScanLine(Row: Integer): Windows.PRGBQUAD;
begin
  if NotEmpty then
    Result := FBits + FSize.cx * (FSize.cy - Row - 1)
  else
    Result := nil;
end;

function TBitmap32.GetHeight: Integer;
begin
  Result := FSize.cy;
end;

function TBitmap32.GetBits: PRGBQUAD;
var
  Data: Windows.DIBSECTION;
begin
  Result := nil;

  if Windows.GetObjectType(Self.FHandle) = OBJ_BITMAP then
  if Windows.GetObject(Self.FHandle, sizeof(Data), @Data) <> 0 then
  begin
    Result := PRGBQUAD(Data.dsBm.bmBits);
  end;
end;

function TBitmap32.GetRect: Classes.TRect;
begin
  Result := Classes.Rect(0, 0, FSize.cx, FSize.cy);
end;

function TBitmap32.GetWidth: Integer;
begin
  Result := FSize.cx;
end;

constructor TBitmap32.Create;
begin
  FCanvas := TCanvas.Create;
  FCanvas.Handle :=
    Windows.CreateCompatibleDC(0);
end;

destructor TBitmap32.Destroy;
begin
  Clear;
  Windows.DeleteDC(FCanvas.Handle);
  FCanvas.Free;
end;

procedure TBitmap32.Clear;
begin
  if FHandle <> 0 then
  begin
    DeleteObject(FHandle);
    FHandle := 0;
  end;
  FBits := nil;
  FSize.cx := 0;
  FSize.cy := 0;
end;

procedure TBitmap32.CreateNew(AWidth, AHeight: Integer; AColor: TColor);
var
  Info: Windows.BITMAPINFO;
  P: Pointer;
begin
  Clear;

  if(AWidth > 0)and(AHeight > 0)then
  begin
    FSize.cx := AWidth;
    FSize.cy := AHeight;

    FillChar(Info, sizeof(Info), 0);
    with Info.bmiHeader do
    begin
      biSize := sizeof(BITMAPINFOHEADER);
      biWidth := AWidth;
      biHeight := AHeight;
      biPlanes := 1;
      biBitCount := 32;
      biCompression := BI_RGB;
    end;
    FHandle := Windows.CreateDIBSection(0, Info, DIB_RGB_COLORS, P, 0, 0);

    FBits := Self.GetBits;
    SelectObject(FCanvas.Handle, FHandle);

    FCanvas.Brush.Color := AColor;
    FCanvas.FillRect(0, 0, FSize.cx, FSize.cy);
  end;
end;

function TBitmap32.Duplicate: TBitmap32;
begin
  Result := TBitmap32.Create;
  Result.CreateNew(FSize.cx, FSize.cy);
  PaintTo(Result.Canvas, GetRect);
end;

procedure TBitmap32.PaintTo(DstCanvas: TCanvas; DstRect: Classes.TRect);
begin
  Windows.StretchBlt(
    DstCanvas.Handle, DstRect.Left, DstRect.Top,
    DstRect.Right - DstRect.Left, DstRect.Bottom - DstRect.Top,
    FCanvas.Handle,
    0, 0, FSize.cx, FSize.cy, SRCCOPY);
end;

procedure TBitmap32.PaintFrom(SrcCanvas: TCanvas; SrcRect: Classes.TRect);
begin
  Windows.StretchBlt(
    FCanvas.Handle,
    0, 0, FSize.cx, FSize.cy,
    SrcCanvas.Handle,
    SrcRect.Left, SrcRect.Top,
    SrcRect.Right - SrcRect.Left, SrcRect.Bottom - SrcRect.Top,
    SRCCOPY);
end;

procedure TBitmap32.SaveAsJPGStream(Stream: TStream;
  CompressionQuality: TJPEGQualityRange);
var
  Data: Windows.DIBSECTION;
  Picture: TPicture;
begin
  if Windows.GetObjectType(Self.FHandle) = OBJ_BITMAP then
  if Windows.GetObject(Self.FHandle, sizeof(Data), @Data) <> 0 then
  begin
    Picture := TPicture.Create;
    try
      Picture.Jpeg.Width := FSize.cx;
      Picture.Jpeg.Height := FSize.cy;
      Picture.Jpeg.Canvas.Brush.Color := clWhite;
      Picture.Jpeg.Canvas.FillRect(GetRect);
      PaintTo(Picture.Jpeg.Canvas, GetRect);

      Picture.Jpeg.CompressionQuality := CompressionQuality;
      Picture.Jpeg.SaveToStream(Stream);
    finally
      Picture.Free;
    end;
  end;
end;

procedure TBitmap32.SaveAsJPGFile(FileName: string;
  CompressionQuality: TJPEGQualityRange);
var
  Stream: LazUTF8Classes.TFileStreamUTF8;
begin
  Stream := LazUTF8Classes.TFileStreamUTF8.Create(FileName, fmCreate);
  try
    SaveAsJPGStream(Stream, CompressionQuality);
  finally
    Stream.Free;
  end;
end;

procedure TBitmap32.LoadFromStream(Stream: TStream);
var
  Picture: TPicture;
begin
  Picture := TPicture.Create;
  try
    Picture.JPEG.LoadFromStream(Stream);
    CreateNew(Picture.Graphic.Width, Picture.Graphic.Height, clWhite);
    FCanvas.Draw(0, 0, Picture.Graphic);
  finally
    Picture.Free;
  end;
end;

procedure TBitmap32.LoadFromFile(FileName: string);
var
  FS: LazUTF8Classes.TFileStreamUTF8;
begin
  FS := LazUTF8Classes.TFileStreamUTF8.Create(FileName, fmOpenRead or fmShareDenyNone);
  try
    LoadFromStream(FS);
  finally
    FS.Free;
  end;
end;

function TBitmap32.GetPixel(X, Y: Integer): PRGBQUAD;
begin
  if NotEmpty then
    Result := ScanLine[Y] + X
  else
    Result := nil;
end;

procedure TBitmap32.SetPixel(X, Y: Integer; R, G, B: BYTE);
var
  P: PBYTE;
begin
  P := PBYTE(GetPixel(X, Y));
  P^ := B;
  Inc(P);
  P^ := G;
  Inc(P);
  P ^ := R;
end;

function TBitmap32.NotEmpty: Boolean;
begin
  Result := (FHandle <> 0)and(FSize.cx > 0)and(FSize.cy > 0);
end;

end.

